#ifndef __BUSBUDDY_H
#define __BUSBUDDY_H
/*

Bus Buddy
=========

Function for calling functions on an external microcontroller.
Using the "V-Tech Bus Ficker"-Cartridge and the "Bus Buddy" sketch running on an Arduino MEGA.

This header file provides functions to communicate with the BusBuddy, issue commands and get results.

Since my "Bus Ficker" cartridge hardware is kinda flaky, I need to employ checksums to correctly (re)transmit data between the V-Tech and Arduino.
(Reason: The ~F0_RD and ~F0_WR on the "Bus Ficker" cartridge Rev.C are not clean and sometimes trigger twice/not at all)

How this file is structured works:
	* bf_xxx functions do the hardware communication with the "Bus Ficker" cartridge (i.e. memory mapped I/O)
	* bb_xxx functions run on top of that, providing high-level access to the "Bus Buddy" Arduino sketch
	* V-Tech sends a "frame" to the memory mapped I/O, this data is parsed (checksum) and handled by the Arduino.
	* The Arduino answers to commands and re-transmits the last answer if so requested (checksum mismatch etc.)
	* If events should be fired, they need to be acively polled as there are no interrupts on the V-Tech side (yet?)

2019-07-27 Bernhard "HotKey" Slawik
*/


// Low-level stuff
#define BF_MMIO_ADDRESS (0x8000 + 0x2000)
//#define BB_DEBUG_TRAFFIC Show each payload character as it comes in frrom the bus

void bf_init() {
	// Switch some high address lines high to activate the BUSFICKER function decoder
	// Caution! "out (3),0x81" will crash the MAME emulation
	__asm
		push af
		
		ld a, #0x81
		out	(3), a
		
		pop af
	__endasm;
}

byte bf_getByte() {
	// Get the byte currently at BusFicker F0_IN
	return *((byte *)(BF_MMIO_ADDRESS));
}
void bf_putByte(byte v) {
	// Output a byte to BusFicker F0_OUT
	*(byte *)BF_MMIO_ADDRESS = v;
}
/*
void bf_put_forceful(byte v) {
	// Don't know the current MMIO address? No problem! :-D
	word aw;
	
	for(aw = 0x8000; aw < 0xc000; aw++) {	// HELLO!!! SOMEONE THERE?!?!? *KNOCK!*KNOCK!*KNOCK!*KNOCK!*
		*(byte *)(aw) = v;
	}
}
*/

// High-Level BusBuddy defines
//const byte BB_BUFFER_SIZE = 64;
#define BB_USE_FILES	// Include everything needed to handle files/SD cards

#define BB_BUFFER_SIZE 128
// Temporary buffer for a frame
byte bb_frame[BB_BUFFER_SIZE];

// Special protocol bytes
const byte BB_PAD = 0x00;
const byte BB_START = '#';
const byte BB_END = '\n';

// Commands
#define BB_CMD_PING 'P'
#define BB_CMD_DEBUG 'D'
#define BB_CMD_RESEND 'R'

#ifdef BB_USE_FILES
	#define BB_CMD_SD_INIT 0x20
	#define BB_CMD_SD_EXISTS 0x21
	#define BB_CMD_SD_OPEN 0x22
	//#define BB_CMD_SD_CLOSE 0x23
	//#define BB_CMD_SD_REMOVE 0x4
	//#define BB_CMD_SD_MKDIR 0x25
	//#define BB_CMD_SD_RMDIR 0x26
	#define BB_CMD_FILE_CLOSE 0x27
	#define BB_CMD_FILE_READ 0x28
	#define BB_CMD_FILE_WRITE 0x29
	//#define BB_CMD_FILE_SEEK 0x2A
	//#define BB_CMD_FILE_SIZE 0x2B
	
	#define BB_HANDLE byte
	#define BB_FILE_READ 0
	#define BB_FILE_WRITE 1
#endif	// BB_USE_FILES

#define bb_putByte(b) bf_putByte(b)
void bb_putWord(word w) {
	bf_putByte(w >> 8);
	bf_putByte(w & 0xff);
}

void bb_putFrame(byte *txData, word txSize) {
	word i;
	byte *d;
	byte b;
	byte check;
	
	// Send txData
	bb_putByte(BB_START);
	
	bb_putWord(txSize);
	
	check = 0x55;
	d = txData;
	for (i = 0; i < txSize; i++) {
		b = *d++;
		bb_putByte(b);
		check ^= b;
	}
	
	bb_putByte(check);
	
	bb_putByte(BB_END);
}

#define bb_getByte bf_getByte
word bb_getWord() {
	return ((word)bb_getByte() << 8) | bb_getByte();
}
int bb_getFrame(byte *rxData, word rxMaxSize) {
	word i;
	word rxSize;
	word timeout;
	byte *d;
	byte b;
	byte checkGiven;
	byte checkActual;
	
	// Receive txData
	
	// Wait for start / sync
	timeout = 0xF000;
	do {
		b = bb_getByte();
		
		// Timeout
		timeout--;
		if (timeout <= 0) {
			printf("Timeout\n");
			return -1;
		}
	} while (b != BB_START);
	
	
	rxSize = bb_getWord();
	
	if (rxSize > rxMaxSize) {
		printf("RX too big (%d > %d)\n", rxSize, rxMaxSize);
		return -2;
	}
	
	#ifdef BB_DEBUG_TRAFFIC
	printf("RX %d: ", rxSize);	// Debugging
	#endif
	
	// Receive data and build checksum
	checkActual = 0x55;
	d = rxData;
	for (i = 0; i < rxSize; i++) {
		b = bb_getByte();
		#ifdef BB_DEBUG_TRAFFIC
		printf("%02X ", b);
		#endif
		*d++ = b;
		checkActual ^= b;
		
	}
	
	#ifdef BB_DEBUG_TRAFFIC
	printf("\n");	// Debugging
	#endif
	
	// Compare checksums
	checkGiven = bb_getByte();
	if (checkGiven != checkActual) {
		printf("Chk mismatch! (%02X != %02X)\n", checkGiven, checkActual);
		return -3;
	}
	
	// Check for end character
	b = bb_getByte();
	if (b != BB_END) {
		printf("No BB_END! (%02X)\n", b);
		return -4;
	}
	
	return rxSize;
}



// BusBuddy Functions
void bb_cmd(byte cmd) {
	// Send a command frame without arguments
	bb_frame[0] = cmd;
	bb_putFrame(&bb_frame[0], 1);
}

void bb_cmd_byte(byte cmd, byte v) {
	// Send a command frame with 1-byte argument
	bb_frame[0] = cmd;
	bb_frame[1] = v;
	bb_putFrame(&bb_frame[0], 2);
}
void bb_cmd_byte2(byte cmd, byte v1, byte v2) {
	// Send a command frame with two 1-byte arguments
	bb_frame[0] = cmd;
	bb_frame[1] = v1;
	bb_frame[2] = v2;
	bb_putFrame(&bb_frame[0], 3);
}
void bb_cmd_asciiz(byte cmd, const char *s) {
	// Send a command frame with zero-terminated string argument
	byte b;
	byte *bs;
	byte *f;
	byte l;
	
	f = &bb_frame[0];
	*f++ = cmd;
	
	//@TODO: memcpy!
	bs = (byte *)s;
	l = 0;
	while(1) {
		b = *bs++;
		if (b == 0) break;	// Do not include zero byte in payload
		*f++ = b;
		l ++;
	}
	
	bb_putFrame(&bb_frame[0], 1+l);
}

void bb_cmd_data(byte cmd, byte *data, const byte dataLen) {
	// Send a command frame with variable length argument
	byte *f;
	byte l;
	
	f = &bb_frame[0];
	*f++ = cmd;
	
	//@TODO: memcpy/strcpy!
	l = dataLen;
	while (l > 0) {
		*f++ = *data++;
		l--;
	}
	bb_putFrame(&bb_frame[0], 1+dataLen);
}
void bb_cmd_byte_data(byte cmd, byte v, byte *data, const byte dataLen) {
	// Send a command frame with a byte plus a variable length argument
	byte *f;
	byte l;
	
	f = &bb_frame[0];
	*f++ = cmd;
	*f++ = v;
	
	//@TODO: memcpy/strcpy!
	l = dataLen;
	while (l > 0) {
		*f++ = *data++;
		l--;
	}
	bb_putFrame(&bb_frame[0], 1+1+dataLen);
}


byte bb_result_byte() {
	// Receive a 1-byte answer. Handles re-sending on error. Returns that byte.
	byte r;
	while (bb_getFrame(&r, 1) < 0) {
		// Error? Ask Arduino to resend last frame
		bb_cmd(BB_CMD_RESEND);
	}
	return r;
}
word bb_result_word() {
	// Receive a 16 bit word answer (MSB first). Handles re-sending on error. Returns that word.
	byte r[2];
	while (bb_getFrame(&r[0], 2) < 0) {
		// Error? Ask Arduino to resend last frame
		bb_cmd(BB_CMD_RESEND);
	}
	return (((word)r[0]) << 8) | r[1];
}

byte bb_result_data(byte *buf, byte lMax) {
	// Receive a variable byte answer. Handles re-sending on error. Returns the actual number of bytes.
	int l;
	
	do {
		l = bb_getFrame(buf, lMax);
		if (l < 0) {
			// Error? Ask Arduino to resend last frame
			bb_cmd(BB_CMD_RESEND);
		}
	} while (l < 0);
	return l;
}


byte bb_ping(byte v) {
	byte r;
	
	//printf("Ping %02X...", v);
	bb_cmd_byte(BB_CMD_PING, v);
	
	r = bb_result_byte();
	//printf("Pong %02X!\n", r);
	
	return r;
}


#ifdef BB_USE_FILES
byte sd_init() {
	byte ok;
	bb_cmd(BB_CMD_SD_INIT);
	ok = bb_result_byte();
	return ok;
}

BB_HANDLE sd_open(const char *filename, byte mode) {
	byte b;
	byte *bs;
	byte *f;
	byte l;
	BB_HANDLE handle;
	
	f = &bb_frame[0];
	*f++ = BB_CMD_SD_OPEN;
	*f++ = mode;
	
	//@TODO: memcpy(dest, src, size) / strcpy(dest, src)!
	bs = (byte *)filename;
	l = 0;
	while(1) {
		b = *bs++;
		if (b == 0) break;
		*f++ = b;
		l ++;
	}
	bb_putFrame(&bb_frame[0], 2+l);
	
	
	//@TODO: What happens if the file does not exist? Negative handle?
	handle = bb_result_byte();
	
	
	return handle;
}

byte sd_exists(const char *filename) {
	byte ex;
	
	bb_cmd_asciiz(BB_CMD_SD_EXISTS, (byte *)filename);
	ex = bb_result_byte();
	return ex;
}

void file_close(BB_HANDLE handle) {
	bb_cmd_byte(BB_CMD_FILE_CLOSE, handle);
	// Has no return value
}

byte file_read(BB_HANDLE handle, byte *buf, byte l) {
	
	bb_cmd_byte2(BB_CMD_FILE_READ, handle, l);
	
	l = bb_result_data(buf, l);
	// Number of bytes read is equal to the size of the payload
	
	return l;
}
void file_write(BB_HANDLE handle, byte *buf, byte l) {
	
	bb_cmd_byte_data(BB_CMD_FILE_WRITE, handle, buf, l);
	
}
#endif	// BB_USE_FILES

#endif //__BUSBUDDY_H