/*

BusFicker Test
==============

Tests for the "V-Tech Bus Ficker"-Cartridge and "Bus Buddy" Arduino sketch.
-> This file is a self-containted "app". If you plan on using "Bus Buddy", please refer to includes/busbuddy.h!

Since my hardware is kinda flaky, I need to employ checksums to correctly (re)transmit data between V-Tech and Arduino.
(Reason: The ~F0_RD and ~F0_WR on the "Bus Ficker" cartridge Rev.C are not clean and sometimes trigger twice/not at all)

How it works:
	* bf_xxx functions do the hardware communication with the "Bus Ficker" cartridge
	* bb_xxx functions run on top of that, providing high-level access to the "Bus Buddy" Arduino sketch
	* V-Tech sends a "frame" to the memory mapped I/O, this data is parsed (checksum) and handled by the Arduino.
	* The Arduino answers to commands and re-transmits the last answer if so requested (checksum mismatch etc.)
	* If events should be fired, they need to be acively polled as there are no interrupts on the V-Tech side (yet?)

2019-07-27 Bernhard "HotKey" Slawik

---

Address lines:
	A 0	0x0001
	A 1	0x0002
	A 2	0x0004
	A 3	0x0008
	A 4	0x0010
	A 5	0x0020
	A 6	0x0040
	A 7	0x0080
	A 8	0x0100
	A 9	0x0200
	A10	0x0400
	A11	0x0800
	A12	0x1000
	A13	0x2000
	A14	0x4000
	A15	0x8000

*/

#include <app.h>	// Load all app defines (to re-use the loader's API)


#define BF_MMIO_ADDRESS (0x8000 + 0x2000)
//#define BB_DEBUG_TRAFFIC Show each payload character as it comes in frrom the bus

void bf_init() {
	// Switch some high address lines high to activate the BUSFICKER function decoder
	__asm
		push af
		
		ld a, #0x81
		out	(3), a
		
		pop af
	__endasm;
}

byte bf_getByte() {
	return *((byte *)(BF_MMIO_ADDRESS));
}
void bf_putByte(byte v) {
	*(byte *)BF_MMIO_ADDRESS = v;
}
/*
void bf_put_forceful(byte v) {
	// Don't know the current MMIO address? No problem! :-D
	word aw;
	
	for(aw = 0x8000; aw < 0xc000; aw++) {	// HELLO!!! SOMEONE THERE?!?!? *KNOCK!*KNOCK!*KNOCK!*KNOCK!*
		*(byte *)(aw) = v;
	}
}
*/

// BusBuddy Arduino sketch
//const byte BB_BUFFER_SIZE = 64;
#define BB_BUFFER_SIZE 64
const byte BB_START = '#';
const byte BB_END = '\n';

#define BB_CMD_PING 'P'
#define BB_CMD_DEBUG 'D'
#define BB_CMD_RESEND 'R'
#define BB_CMD_SD_INIT 0x20
#define BB_CMD_SD_EXISTS 0x21
#define BB_CMD_SD_OPEN 0x22
//#define BB_CMD_SD_CLOSE 0x23
//#define BB_CMD_SD_REMOVE 0x4
//#define BB_CMD_SD_MKDIR 0x25
//#define BB_CMD_SD_RMDIR 0x26
#define BB_CMD_FILE_CLOSE 0x27
#define BB_CMD_FILE_READ 0x28
#define BB_CMD_FILE_WRITE 0x29
//#define BB_CMD_FILE_SEEK 0x2A
//#define BB_CMD_FILE_SIZE 0x2B


#define BB_FILE_READ 0
#define BB_FILE_WRITE 1


#define bb_putByte(b) bf_putByte(b)
void bb_putWord(word w) {
	bf_putByte(w >> 8);
	bf_putByte(w & 0xff);
}

void bb_putFrame(byte *txData, word txSize) {
	word i;
	byte *d;
	byte b;
	byte check;
	
	// Send txData
	bb_putByte(BB_START);
	
	bb_putWord(txSize);
	
	check = 0x55;
	d = txData;
	for (i = 0; i < txSize; i++) {
		b = *d++;
		bb_putByte(b);
		check ^= b;
	}
	
	bb_putByte(check);
	
	bb_putByte(BB_END);
}

#define bb_getByte bf_getByte
word bb_getWord() {
	return ((word)bb_getByte() << 8) | bb_getByte();
}
int bb_getFrame(byte *rxData, word rxMaxSize) {
	word i;
	word rxSize;
	word timeout;
	byte *d;
	byte b;
	byte checkGiven;
	byte checkActual;
	
	// Receive txData
	
	// Wait for start / sync
	timeout = 0xF000;
	do {
		b = bb_getByte();
		
		// Timeout
		timeout--;
		if (timeout <= 0) {
			printf("Timeout\n");
			return -1;
		}
	} while (b != BB_START);
	
	
	rxSize = bb_getWord();
	
	if (rxSize > rxMaxSize) {
		printf("RX too big (%d > %d)\n", rxSize, rxMaxSize);
		return -2;
	}
	
	#ifdef BB_DEBUG_TRAFFIC
	printf("RX %d: ", rxSize);	// Debugging
	#endif
	
	// Receive data and build checksum
	checkActual = 0x55;
	d = rxData;
	for (i = 0; i < rxSize; i++) {
		b = bb_getByte();
		#ifdef BB_DEBUG_TRAFFIC
		printf("%02X ", b);
		#endif
		*d++ = b;
		checkActual ^= b;
		
	}
	
	#ifdef BB_DEBUG_TRAFFIC
	printf("\n");	// Debugging
	#endif
	
	// Compare checksums
	checkGiven = bb_getByte();
	if (checkGiven != checkActual) {
		printf("Chk mismatch! (%02X != %02X)\n", checkGiven, checkActual);
		return -3;
	}
	
	// Check for end character
	b = bb_getByte();
	if (b != BB_END) {
		printf("No BB_END! (%02X)\n", b);
		return -4;
	}
	
	return rxSize;
}


// Temporary buffer for a frame
byte bb_frame[BB_BUFFER_SIZE];

// BusBuddy Functions
void bb_cmd(byte cmd) {
	// Send a command frame without arguments
	bb_frame[0] = cmd;
	bb_putFrame(&bb_frame[0], 1);
}

void bb_cmd_byte(byte cmd, byte v) {
	// Send a command frame with 1-byte argument
	bb_frame[0] = cmd;
	bb_frame[1] = v;
	bb_putFrame(&bb_frame[0], 2);
}
void bb_cmd_byte2(byte cmd, byte v1, byte v2) {
	// Send a command frame with two 1-byte arguments
	bb_frame[0] = cmd;
	bb_frame[1] = v1;
	bb_frame[2] = v2;
	bb_putFrame(&bb_frame[0], 3);
}
void bb_cmd_asciiz(byte cmd, const char *s) {
	// Send a command frame with zero-terminated string argument
	byte b;
	byte *bs;
	byte *f;
	byte l;
	
	f = &bb_frame[0];
	*f++ = cmd;
	
	//@TODO: memcpy!
	bs = (byte *)s;
	l = 0;
	while(1) {
		b = *bs++;
		if (b == 0) break;	// Do not include zero byte in payload
		*f++ = b;
		l ++;
	}
	
	bb_putFrame(&bb_frame[0], 1+l);
}

void bb_cmd_data(byte cmd, byte *data, const byte dataLen) {
	// Send a command frame with variable length argument
	byte *f;
	byte l;
	
	f = &bb_frame[0];
	*f++ = cmd;
	
	//@TODO: memcpy/strcpy!
	l = dataLen;
	while (l > 0) {
		*f++ = *data++;
		l--;
	}
	bb_putFrame(&bb_frame[0], 1+dataLen);
}


byte bb_result_byte() {
	// Receive a 1-byte answer. Handles re-sending on error. Returns that byte.
	byte r;
	while (bb_getFrame(&r, 1) < 0) {
		// Error? Ask Arduino to resend last frame
		bb_cmd(BB_CMD_RESEND);
	}
	return r;
}
byte bb_result_word() {
	// Receive a 16 bit word answer (MSB first). Handles re-sending on error. Returns that word.
	byte r[2];
	while (bb_getFrame(&r, 2) < 0) {
		// Error? Ask Arduino to resend last frame
		bb_cmd(BB_CMD_RESEND);
	}
	return (((word)r[0]) << 8) | r[1];
}

byte bb_result_data(byte *buf, byte lMax) {
	// Receive a variable byte answer. Handles re-sending on error. Returns the actual number of bytes.
	int l;
	
	do {
		l = bb_getFrame(buf, lMax);
		if (l < 0) {
			// Error? Ask Arduino to resend last frame
			bb_cmd(BB_CMD_RESEND);
		}
	} while (l < 0);
	return l;
}


byte bb_ping(byte v) {
	byte r;
	
	//printf("Ping %02X...", v);
	bb_cmd_byte(BB_CMD_PING, v);
	
	r = bb_result_byte();
	//printf("Pong %02X!\n", r);
	
	return r;
}

byte sd_init() {
	byte ok;
	bb_cmd(BB_CMD_SD_INIT);
	ok = bb_result_byte();
	return ok;
}


byte sd_open(const char *filename, byte mode) {
	byte b;
	byte *bs;
	byte *f;
	byte l;
	byte handle;
	
	f = &bb_frame[0];
	*f++ = BB_CMD_SD_OPEN;
	*f++ = mode;
	
	//@TODO: memcpy!
	bs = (byte *)filename;
	l = 0;
	while(1) {
		b = *bs++;
		if (b == 0) break;
		*f++ = b;
		l ++;
	}
	bb_putFrame(&bb_frame[0], 2+l);
	
	
	handle = bb_result_byte();
	return handle;
}

byte sd_exists(const char *filename) {
	byte ex;
	
	bb_cmd_asciiz(BB_CMD_SD_EXISTS, (byte *)filename);
	ex = bb_result_byte();
	return ex;
}

void file_close(byte h) {
	bb_cmd_byte(BB_CMD_FILE_CLOSE, h);
	// Has no return value
}

byte file_read(byte h, byte *buf, byte l) {
	
	bb_cmd_byte2(BB_CMD_FILE_READ, h, l);
	l = bb_result_data(buf, l);
	
	return l;
}
void file_write(byte h, byte *buf, byte l) {
	
	//@TODO: Implement!
	(void)h;
	(void)buf;
	(void)l;
	
	//bb_cmd_data(BB_CMD_FILE_WRITE, buf, l);
	printf("Not implemented!\n");
}


void idle() {
	word t;
	for(t = 0; t < 0x1000; t++) {
		;
	}
}

int main(int argc, char *argv[]) {
	char c;
	byte b;
	word i;
	word timeout;
	byte r;
	byte handle;
	byte l;
	byte tmpData[32];
	
	(void)argc;	(void)argv;	// Mark args as "not used" to suppress warnings
	
	
	printf("BusBuddy Test\n");
	
	bf_init();
	
	i = 0;
	do {
		printf(">");
		c = getchar();
		printf("%s\n", c);
		
		switch(c) {
			case 'p':
				// Ping
				b = i & 0xff;
				printf("Ping %02X...", b);
				b = bb_ping(b);
				printf("Pong %02X!\n", b);
				
				i++;
				break;
			
			case 's':
				printf("Init SD...");
				r = sd_init();
				if (r == 1)	printf("OK\n");
				else		printf("E=%d\n", r);
				
				getchar();
				if (r != 1) continue;
				
				
				printf("Open file...");
				handle = sd_open("test.txt", BB_FILE_READ);
				printf("OK\n");
				
				getchar();
			
				printf("Read...");
				l = file_read(handle, &tmpData[0], 32);
				printf("%d bytes\n", l);
				
				getchar();
				
				printf("\"%s\"\n", (char *)&tmpData[0]);
				getchar();
			
				printf("Close file...");
				file_close(handle);
				printf("OK\n");
				break;
			
			
			case 'd':
				bb_frame[0] = BB_CMD_DEBUG;
				bb_frame[1] = 'H';
				bb_frame[2] = 'e';
				bb_frame[3] = 'l';
				bb_frame[4] = 'l';
				bb_frame[5] = 'o';
				bb_frame[6] = '!';
				bb_putFrame(&bb_frame[0], 7);
				break;
			
			case 'h':
				printf(">>HELO...");
				bf_putByte('H');
				bf_putByte('E');
				bf_putByte('L');
				bf_putByte('O');
				printf("\n");
				break;
			
			case 'c':
				// Wait for any data
				printf("<");
				timeout = 200;
				do {
					b = bf_getByte();
					idle();
					timeout--;
				} while ((b == 0) && (timeout > 0));
				
				if (timeout <= 0) {
					printf("Timeout!\n");
					continue;
				}
				
				// Get all the data
				printf("<");
				putchar(b);
				
				while(b != 0) {
					b = bf_getByte();
					putchar(b);
				}
				printf("\n");
				/*
				printf(">>");
				do {
					c = getchar();
					bf_put(c);
				} while(c != KEY_BREAK);
				printf("---- END ----\n");
				*/
				
				break;
			
			case 'a':
				bf_init();
				
				while(1) {
					printf("FF");
					bf_putByte(0xff);
					bf_putByte(0xff);
					bf_putByte(0xff);
					//bf_put_forceful(0xff);
					for(i = 0; i < 0x8000; i++) {;}
					
					printf("00");
					bf_putByte(0x00);
					bf_putByte(0x00);
					bf_putByte(0x00);
					//bf_put_forceful(0x00);
					for(i = 0; i < 0x8000; i++) {;}
				}
				break;
			
			case 'i':
				bf_init();
				
				while(1) {
					b = bf_getByte();
					printf("%02X", b);
				}
				break;
				
			case 'q':
				// Quit
				break;
			
			default:
				printf("?\n");
		}
		
	} while(c != 'q');
	
	
	return 0;
}


