#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
Serial uploader for V-Tech SDCC serloader.c

2019-07-13 Bernhard "HotKey" Slawik
"""

import serial
import time
import sys	# for sys.argv and sys.exit(status)
import getopt

# FTDI
SERIAL_PORT = '/dev/ttyUSB0'
SERIAL_BAUD = 9600	# If connected directly to the parallel port

# Arduino MEGA
#SERIAL_PORT = '/dev/ttyACM0'
#SERIAL_BAUD = 115200	# If connected to an Arduino which buffers the serial stream





DEFAULT_BIN_FILENAME = 'out/app_parallelBuddy.app.bin'
#DEFAULT_BIN_FILENAME = 'out/app_hello.app.bin'
#DEFAULT_BIN_FILENAME = 'out/app_test.app.bin'

DEFAULT_WAIT_FOR_BOOTLOADER = not True	# Before uploading: Wait for #BLM
DEFAULT_WAIT_FOR_READY = not True	# Before uploading: Wait for #RDY
DEFAULT_WAIT_FOR_RESULT = not True	# Stay connected until program ends

ADDR_TEMP = 0xc400	# Address to temporarily store the app
ADDR_CODE = 0xc400	# Address to re-locate code to
ADDR_DATA = 0xc200	# Address to re-locate data to

# How much bytes to put into a HEX request
# More than 32 requires alteration to the SERIAL_RX_BUFFER_SIZE in Arduino's HardwareSerial.h
#MAX_HEX_PER_LINE = 64
#MAX_HEX_PER_LINE = 60
MAX_HEX_PER_LINE = 48
#MAX_HEX_PER_LINE = 32	# 32 works fine
#MAX_HEX_PER_LINE = 16

PUT_TIMEOUT = 0.5
PUT_THROTTLE = 0.15
PRE_GET_THROTTLE = 0	#0.1
GET_TIMEOUT = 0.5
#GET_THROTTLE = 0.05
PREFIX = ' '	# It is OK to put bytes before the initial '#' character to help the SoftSerial to sync up

def put(txt):
	print('serloader: %s' % txt)

def str_hex(s):
	"Safe string conversion (encoding invalid chars to hex numbers)"
	if type(s) not in [str, unicode]:
		s = str(s)
	
	r = ''
	for b in s:
		c = ord(b)
		if (c < 128): r += chr(c)
		else:
			r += ' [0x%02X] ' % c
	return r

class Uploader:
	def __init__(self, port=SERIAL_PORT):
		self.running = False
		
		self.port = port
		self.ser = None
		
		# Open at start-up...
		self.is_open = self.open()
		
		if not self.is_open:
			put('The port "%s" could not be opened!' % (self.port))
		
	
	def __del__(self):
		self.close()
		pass
	
	def open(self):
		self.ser = None
		port = self.port
		put('Opening "%s"...' % port)
		
		try:
			self.ser = serial.Serial(port=port, baudrate=SERIAL_BAUD, bytesize=8, parity='N', stopbits=1, timeout=3, xonxoff=0, rtscts=0)
			put('Open!')
			return True
		except serial.serialutil.SerialException as e:
			#except e:
			put('Error opening serial device: %s' % str(e))
			return False
	
	def close(self):
		self.running = False
		time.sleep(0.2)
		
		if (self.ser is not None):
			self.ser.close()
			self.ser = None
	
	def write(self, data):
		if self.ser is None:
			put('! Cannot send data, because serial is not open!')
			return False
		
		put('>>> "%s"' % str_hex(data.strip()))
		self.ser.write(data)
		#self.ser.flush()	#?
	
	def readline(self, timeout=1):
		r = ''
		start_time = time.time()
		end_time = start_time + timeout
		data = ''
		while (time.time() < end_time):
			if self.ser.in_waiting > 0:
				"""
				c = self.ser.read(1024)
				put('<<< %d bytes: %s' % (len(c), str(c)))
				data += c
				
				if ('\r' in c) or ('\n' in c):
					data = data.strip()
					put('<<< "%s"' % str_hex(data))
					return data
				"""
				data = self.ser.readline().strip()
				put('<<< "%s"' % str_hex(data))
				
				# Weird... I get a lot of 0xFF....
				#while (len(data) > 0) and (ord(data[0]) == 0xff):
				#	data = data[1:]
				#put('<<< "%s"' % str_hex(data))
				
				return data
				#end_time = time.time() + timeout	# Extend timeout
			else:
				time.sleep(0.1)
		put('Timeout!')
		return None
	
	
	def get_loop(self):
		timeout = 5
		start_time = time.time()
		end_time = start_time + timeout
		while (time.time() < end_time):
			if self.ser.in_waiting:
				while (self.ser.in_waiting > 0):	# inWaiting()
					data = self.ser.readline().strip()
					put('<<< "%s"' % str_hex(data))
				end_time = time.time() + timeout	# Extend timeout
			else:
				time.sleep(0.1)
		put('Timeout!')
	
	
	def wait_for_ready(self):
		### Wait for READY signal
		l = None
		while (l != '#RDY'):
			put('Waiting for ready signal...')
			l = self.readline(timeout=3)
		put('Serial loader is ready!')
	
	def wait_for_bootloader(self):
		### Wait for BLM signal
		l = None
		while (l != '#BLM'):
			put('Waiting for Bootloader...')
			l = self.readline(timeout=3)
		put('Device is in bootloader mode!')
		
	def upload(self, app_filename, wait_for_bootloader=False, wait_for_ready=True, wait_for_result=True):
		
		
		if wait_for_bootloader:
			self.wait_for_bootloader()
			
		if wait_for_ready:
			self.wait_for_ready()
		
		put('Flushing...')
		self.readline(timeout=PUT_THROTTLE)
		
		
		#self.write('#PING')
		with open(app_filename, 'rb') as h:
			data = h.read()
		
		addr_start = ADDR_TEMP
		
		data_len = len(data)
		addr = addr_start
		data_ofs = 0
		while (data_ofs < data_len):
			put('Transmitting %d/%d at 0x%04X...' % (data_ofs, data_len, addr))
			
			n = data_len - data_ofs
			if (n > MAX_HEX_PER_LINE): n = MAX_HEX_PER_LINE	# Only n values at once
			
			# Prepare line
			r = '%s#PUT%02X%04X' % (PREFIX, n, addr)
			for i in range(n):
				r += '%02X' % ord(data[data_ofs + i])
			r += '\n'
			
			verified = False
			while not verified:
				
				# Send and wait for "#POK" reply
				sent = False
				while not sent:
					# Send "#PUT"
					self.write(r)
					
					# Wait for "#POK"
					l = self.readline(timeout=PUT_TIMEOUT)
					if (l == '#POK'):
						sent = True
					else:
						# Throttle
						#pass
						#self.write('\n')	# Flush
						time.sleep(PUT_THROTTLE)
						#time.sleep(0.2)
					
				
				# Verify: Get the memory region and compare
				sent = False
				while not sent:
					
					# Request memory region
					if (PRE_GET_THROTTLE > 0): time.sleep(PRE_GET_THROTTLE)
					
					self.write('%s#GET%02X%04X\n' % (PREFIX, n, addr))
					
					# Flush
					
					l = self.readline(timeout=GET_TIMEOUT)
					# Compare
					if (l is not None) and (len(l) >= (4+4+n*2)):
						# Size matches
						sent = True
						
						try:
							identical = True
							for i in range(n):
								b = l[4+4 + 2*i:4+4 + 2*i+2]
								vd = int(b, 16)
								vs = ord(data[data_ofs + i])
								if (vs != vd):
									put('Byte mismatch at %d: Should be %02X, but it is %02X!' % (i, vs, vd))
									identical = False
									break;
						except ValueError:
							identical = False
						
						if identical:
							# Okay!
							#put('Verify OK!')
							verified = True
						else:
							# Throttle
							time.sleep(PUT_THROTTLE)
					else:
						# Timeout or size mismatch
						#put('Size mismatch')
						sent = False
						
						# Throttle
						#time.sleep(GET_THROTTLE)
						#self.write('\n')
						time.sleep(PUT_THROTTLE)
					#
				#
			#
			
			# Next chunk
			addr += n
			data_ofs += n
			
		#
		put('Transmitted.')
		
		# Throttle
		time.sleep(PUT_THROTTLE)
		
		put('Running 0x%04X...' % (addr_start))
		
		# Run it (and wait for "#ROK")
		sent = False
		while not sent:
			# Send command to re-locate and run
			self.write('%s#RUN%04X%04X%04X\n' % (PREFIX, addr_start, ADDR_CODE, ADDR_DATA))
			l = self.readline(timeout=1)
			if (l is not None) and (l.startswith('#ROK')):
				sent = True
			else:
				time.sleep(PUT_THROTTLE)
		
		put('Running!')
		
		if wait_for_result:
			# Wait for "#RET"
			r = None
			finished = False
			while not finished:
				l = self.readline(timeout=10)
				if (l is not None) and (l.startswith('#RET')):
					r = int(l[4:8], 16)
					finished = True
			put('Finished. r=%s' % str(r))
			
			sys.exit(r)
		
		else:
			put('Flushing...')
			self.readline()
		
		return
	
	


def show_help():
	put(__doc__)
	put('serloader_uploader.py -i <app file>')

if __name__ == '__main__':
	#put(__doc__)
	
	argv = sys.argv[1:]
	try:
		opts, args = getopt.getopt(argv, 'hni:', ['input=','no-result'])
	except getopt.GetoptError:
		show_help()
		sys.exit(2)
	
	bin_filename = DEFAULT_BIN_FILENAME
	
	wait_for_bootloader = DEFAULT_WAIT_FOR_BOOTLOADER
	wait_for_ready = DEFAULT_WAIT_FOR_READY
	wait_for_result = DEFAULT_WAIT_FOR_RESULT
	
	for opt, arg in opts:
		if opt == '-h':
			show_help()
			sys.exit()
		elif opt in ('-n', '--no-result'):
			wait_for_result = False
		elif opt in ('-i', '--input'):
			bin_filename = arg
	
	u = Uploader()
	
	if not u.is_open:
		put('Serial port could not be opened. Aborting.')
		sys.exit(4)
	
	put('Uploading app binary "%s"...' % (bin_filename))
	u.upload(bin_filename, wait_for_bootloader, wait_for_ready, wait_for_result)
	
	
	put('End.')
